import argparse
import dataclasses
import os
import shutil

from ._buddy_cli_subcommand import Subcommand
from ._buddy_cli_utils import BuddyPaths

_TRASH_DIR = "./_trash/"


def _delete(path, forever) -> None:
    assert os.path.exists(path)

    if forever:
        # Delete file/directory forever
        if os.path.isdir(path):
            print(f"Deleting {path} (recursive)")
            shutil.rmtree(path)
        elif os.path.isfile(path):
            print(f"Deleting {path}")
            os.remove(path)
        else:
            assert False, "Something went wrong"
    else:
        # Move files/directory to a new path
        new_path = os.path.join(_TRASH_DIR, path)

        # Create trash directory if it doesn't exist yet
        directory = os.path.dirname(new_path)
        if not os.path.isdir(directory):
            os.makedirs(directory)

        # Move file/directory to trash
        print(f"Moving {path} to {new_path}")
        os.rename(path, new_path)


class DeleteSubcommand(Subcommand):
    """Delete a Buddy experiment.
    """

    subcommand: str = "delete"

    @classmethod
    def add_arguments(
        cls, *, parser: argparse.ArgumentParser, paths: BuddyPaths
    ) -> None:
        parser.add_argument(
            "experiment_name",
            type=str,
            help="Name of experiment, as printed by `$ buddy list`.",
            metavar="experiment_name",  # Set metavar => don't show choices in help menu
            choices=paths.find_experiments().experiment_names,
        )
        parser.add_argument(
            "--forever",
            action="store_true",
            help=f"Delete experiment forever: if unset, move files into `{_TRASH_DIR}`.",
        )

    @classmethod
    def main(cls, *, args: argparse.Namespace, paths: BuddyPaths) -> None:
        # Get experiment name
        experiment_name = args.experiment_name

        # If we're just moving an experiment, check that it doesn't exist already
        if not args.forever:
            new_checkpoint_files = dataclasses.replace(
                paths, checkpoint_dir=os.path.join(_TRASH_DIR, paths.checkpoint_dir)
            ).find_checkpoints(experiment_name)
            if len(new_checkpoint_files) != 0:
                raise RuntimeError(
                    "Checkpoints for matching experiment name already exist in trash; "
                    "rename experiment before deleting."
                )
            if os.path.exists(
                os.path.join(_TRASH_DIR, paths.log_dir, f"{experiment_name}")
            ):
                raise RuntimeError(
                    "Logs for matching experiment name already exist in trash; "
                    "rename experiment before deleting."
                )
            if os.path.exists(
                os.path.join(_TRASH_DIR, paths.metadata_dir, f"{experiment_name}.yaml")
            ):
                raise RuntimeError(
                    "Metadata for matching experiment name already exist in trash; "
                    "rename experiment before deleting."
                )

        # Delete checkpoint files
        checkpoint_paths = paths.find_checkpoints(experiment_name)
        print(f"Found {len(checkpoint_paths)} checkpoint files")
        for path in checkpoint_paths:
            _delete(path, args.forever)

        # Delete metadata
        metadata_path = os.path.join(paths.metadata_dir, f"{experiment_name}.yaml")
        if os.path.exists(metadata_path):
            _delete(metadata_path, args.forever)
        else:
            print("No metadata found")

        # Delete logs
        log_path = os.path.join(paths.log_dir, f"{experiment_name}")
        if os.path.exists(log_path):
            _delete(log_path, args.forever)
        else:
            print("No logs found")



import abc
import argparse

from ._buddy_cli_utils import BuddyPaths


class Subcommand(abc.ABC):
    """Subcommand interface: defines arguments, runtime routine.
    """

    subcommand: str
    helptext: str

    @classmethod
    @abc.abstractclassmethod
    def add_arguments(
        cls, *, parser: argparse.ArgumentParser, paths: BuddyPaths
    ) -> None:
        raise NotImplementedError

    @classmethod
    @abc.abstractclassmethod
    def main(cls, *, args: argparse.Namespace, paths: BuddyPaths) -> None:
        raise NotImplementedError



import numpy as np
import torch

import fannypack.utils


def test_quadratic_matmul():
    """Tests quadratic_matmul() by checking its backward pass.
    """
    N = 100
    D = 5

    # Deterministic test
    torch.random.manual_seed(0)

    # Compute quadratic
    x = torch.randn((N, D), requires_grad=True)
    A = torch.randn((N, D, D))
    xTAx = fannypack.utils.quadratic_matmul(x, A)

    # Check shape
    assert xTAx.shape == (N,)

    # Check gradient
    torch.sum(xTAx).backward()
    np.testing.assert_allclose(
        *fannypack.utils.to_numpy(
            [x.grad, ((A.transpose(-1, -2) + A) @ x[:, :, None]).squeeze(),]
        ),
        rtol=1e-4,
        atol=1e-6,
    )


def test_gaussian_log_prob():
    """Check that our Gaussian log probability implementation matches the native PyTorch
    one.
    """
    N = 100
    D = 5

    # Deterministic test
    torch.random.manual_seed(0)

    # Compute quadratic
    mean = torch.randn((N, D))
    value = torch.randn((N, D))
    covariance = torch.randn((N, D, D))
    covariance = covariance @ covariance.transpose(-1, -2)

    # Check output
    np.testing.assert_allclose(
        *fannypack.utils.to_numpy(
            [
                torch.distributions.MultivariateNormal(
                    loc=mean, covariance_matrix=covariance
                ).log_prob(value=value),
                fannypack.utils.gaussian_log_prob(
                    mean=mean, covariance=covariance, value=value,
                ),
            ]
        ),
        rtol=1e-3,
        atol=1e10,
    )


def test_tril_count_simple():
    """Basic counting utility check.
    """
    assert fannypack.utils.matrix_dim_from_tril_count(6) == 3
    assert fannypack.utils.tril_count_from_matrix_dim(3) == 6


def test_tril_count_bijective():
    """Check bijectivity of our counting utilities.
    """

    for matrix_dim in range(1, 100):
        assert (
            fannypack.utils.matrix_dim_from_tril_count(
                fannypack.utils.tril_count_from_matrix_dim(matrix_dim)
            )
            == matrix_dim
        )


def test_tril_vector_conversions():
    """Check that our `tril_to_vector` and `vector_to_tril` functions return the correct
    shapes + bijectivity.
    """

    for batch_dims in ((), (3, 2, 5), (7,)):
        for matrix_dim in range(1, 10):
            tril_count = fannypack.utils.tril_count_from_matrix_dim(matrix_dim)
            vectors = torch.randn(batch_dims + (tril_count,))
            trils = fannypack.utils.tril_from_vector(vectors)

            assert trils.shape == batch_dims + (matrix_dim, matrix_dim)
            assert torch.all(fannypack.utils.vector_from_tril(trils) == vectors)
